import json
import os
from pathlib import Path
from typing import Dict, List


def load_seed(file):
    current_folder = os.path.dirname(__file__)
    data = Path(os.path.join(current_folder, file)).read_text()
    return json.loads(data)


def mapbit(length: int, trues: List = [], falses: List = []):
    def _value(key: int):
        if trues:
            return key in trues
        if falses:
            return key not in falses
        return True

    return {i + 1: _value(i + 1) for i in range(length)}


def mapbit_from_dict(length: int, trues: Dict) -> Dict[int, bool]:
    trues = [int(i) for i, _ in trues.items()]
    falses = [i + 1 for i in range(length)]
    return mapbit(length or len(trues), trues=trues, falses=falses)



from typing import Dict


class PubTCEntry:

    """Represents a Publisher TC entry that can be used to retrieve further information.

    :param purposes_consent: The publisher's purposes consent.
    :param purposes_lit_transparency: The publisher's purposes legitimate
        interest consents.
    :param custom_purposes_consent: The publisher's custom purposes consent.
    :param custom_purposes_lit_transparency: The publisher's custom purposes
        legitimate interest consents.
    """

    def __init__(
        self,
        purposes_consent: Dict[int, bool],
        purposes_lit_transparency: Dict[int, bool],
        custom_purposes_consent: Dict[int, bool],
        custom_purposes_lit_transparency: Dict[int, bool],
    ):
        self.purposes_consent = purposes_consent
        self.purposes_lit_transparency = purposes_lit_transparency
        self.custom_purposes_consent = custom_purposes_consent
        self.custom_purposes_lit_transparency = custom_purposes_lit_transparency



from typing import List

from .bits import Reader
from .iab_tcf import base64_decode, segments
from .v2.non_core_segments import NonCoreSegment
from .v2.publisher_restriction import PubRestrictionEntry


class ConsentV2:

    """Represents a v2 consent with all the information extracted.

    :param consent: The consent to process in bytes.
    """

    def __init__(self, consent: bytes):
        self._reader: Reader = Reader(consent)
        self.version = self._reader.read_int(6)
        self.created = self._reader.read_time()
        self.last_updated = self._reader.read_time()
        self.cmp_id = self._reader.read_int(12)
        self.cmp_version = self._reader.read_int(12)
        self.consent_screen = self._reader.read_int(6)
        self.consent_language = self._reader.read_string(2)
        self.vendor_list_version = self._reader.read_int(12)
        self.tcf_policy_version = self._reader.read_int(6)
        self.is_service_specific = self._reader.read_bool()
        self.use_non_standard_stacks = self._reader.read_bool()
        self.special_features_optin = self._reader.read_bitfield(12)
        self.purposes_consent = self._reader.read_bitfield(24)
        self.purposes_legitimate_interests = self._reader.read_bitfield(24)
        self.purpose_one_treatment = self._reader.read_bool()
        self.publisher_cc = self._reader.read_string(2)
        self.read_consent_vendors()
        self.read_interest_vendors()
        self.read_pub_restriction_entries()

    def read_consent_vendors(self):
        """Reads the consent vendors. It must be called with the
        reader already in the position where the consent vendors start.
        """
        self.max_consent_vendor_id = self._reader.read_int(16)
        self.is_consent_range_encoding = self._reader.read_bool()
        if self.is_consent_range_encoding:
            self.num_consent_entries = self._reader.read_int(12)
            self.consented_vendors_range = self._reader.read_range(
                self.num_consent_entries
            )
        else:
            self.consented_vendors = self._reader.read_bitfield(
                self.max_consent_vendor_id
            )

    def read_interest_vendors(self):
        """Reads the interest vendors. It must be called with the
        reader already in the position where the interest vendors start.
        """
        self.max_interests_vendor_id = self._reader.read_int(16)
        self.is_interests_range_encoding = self._reader.read_bool()
        if self.is_interests_range_encoding:
            self.num_interests_entries = self._reader.read_int(12)
            self.interests_vendors_range = self._reader.read_range(
                self.num_interests_entries
            )
        else:
            self.interests_vendors = self._reader.read_bitfield(
                self.max_interests_vendor_id
            )

    def read_pub_restriction_entries(self):
        """Reads the publisher restriction entries. It must be called with the
        reader already in the position where the publisher restrictions start.
        """
        self.num_pub_restrictions = self._reader.read_int(12)
        self.pub_restriction_entries = []
        for _ in range(self.num_pub_restrictions):
            purpose_id = self._reader.read_int(6)
            restriction_type = self._reader.read_int(2)
            num_entries = self._reader.read_int(12)
            restrictions_range = self._reader.read_range(num_entries)
            self.pub_restriction_entries.append(
                PubRestrictionEntry(
                    purpose_id=purpose_id,
                    restriction_type=restriction_type,
                    restrictions_range=restrictions_range,
                )
            )

    def read_non_core_segments(self, segments: List[str]):
        """Receives list of non core segments and tries to
        parse the information inside them.
        """
        for segment in segments[1:]:
            decoded = base64_decode(segment)
            non_core_segment = NonCoreSegment(decoded)
            if non_core_segment.is_disclosed_vendors():
                self.oob_disclosed_vendors = non_core_segment.read_vendors()
            elif non_core_segment.is_allowed_vendors():
                self.oob_allowed_vendors = non_core_segment.read_vendors()
            elif non_core_segment.is_publisher_tc():
                self.publisher_tc = non_core_segment.read_publisher_tc()

    def is_purpose_allowed(self, id: int) -> bool:
        """Checks if a purpose is allowed or not.

        :param id: Purpose id to check if it's allowed or not.
        """
        if id in self.purposes_consent:
            return self.purposes_consent[id]
        return False

    def has_purpose_legitimate_interest(self, id: int) -> bool:
        if id in self.purposes_legitimate_interests:
            return self.purposes_legitimate_interests[id]
        return False

    def is_vendor_allowed(self, id: int) -> bool:
        """Checks if a vendor is allowed or not.

        :param id: Vendor id to check if it's allowed or not.
        """
        if self.is_consent_range_encoding:
            for (start, end) in self.consented_vendors_range:
                if start <= id and id <= end:
                    return True
            return False
        return False if id not in self.consented_vendors else self.consented_vendors[id]

    def is_interest_allowed(self, id: int) -> bool:
        if self.is_interests_range_encoding:
            for (start, end) in self.interests_vendors_range:
                if start <= id and id <= end:
                    return True
            return False
        return False if id not in self.interests_vendors else self.interests_vendors[id]

    def get_restriction(self, publisher: int, purpose: int) -> PubRestrictionEntry:
        for entry in self.pub_restriction_entries:
            if entry.purpose_id == purpose and entry.is_in_range(publisher):
                return entry
        return None


def decode_v2(consent: str):
    """Decodes a v2 consent string that it's encoded in base64 but split in
    segments.

    :param consent: base64 encoded consent string.
    """
    consent_segments = segments(consent)
    consent = ConsentV2(base64_decode(consent_segments[0]))
    consent.read_non_core_segments(consent_segments)
    return consent



from .bits import Reader
from .iab_tcf import base64_decode


class ConsentV1:

    """Represents a v1.1 consent with all the information extracted.

    :param consent: The consent to process in bytes.
    """

    def __init__(self, consent: bytes):
        self._reader: Reader = Reader(consent)
        self.version = self._reader.read_int(6)
        self.created = self._reader.read_time()
        self.last_updated = self._reader.read_time()
        self.cmp_id = self._reader.read_int(12)
        self.cmp_version = self._reader.read_int(12)
        self.consent_screen = self._reader.read_int(6)
        self.consent_language = self._reader.read_string(2)
        self.vendor_list_version = self._reader.read_int(12)
        self.purposes_allowed = self._reader.read_bitfield(24)
        self.max_vendor_id = self._reader.read_int(16)
        self.is_range_encoding = self._reader.read_bool()
        if self.is_range_encoding:
            self.default_consent = self._reader.read_bool()
            self.num_entries = self._reader.read_int(12)
            self.range_entries = self._reader.read_range(self.num_entries)
        else:
            self.consented_vendors = self._reader.read_bitfield(self.max_vendor_id)

    def is_purpose_allowed(self, id: int) -> bool:
        """Checks if a purpose is allowed or not.

        :param id: Purpose id to check if it's allowed or not.
        """
        if id in self.purposes_allowed:
            return self.purposes_allowed[id]
        return False

    def is_vendor_allowed(self, id: int) -> bool:
        """Checks if a vendor is allowed or not.

        :param id: Vendor id to check if it's allowed or not.
        """
        if self.is_range_encoding:
            for (start, end) in self.range_entries:
                if start <= id and id <= end:
                    return not self.default_consent
            return self.default_consent
        return False if id not in self.consented_vendors else self.consented_vendors[id]


def decode_v1(consent: str):
    """Decodes a v1.1 consent string that it's encoded in base64.

    :param consent: base64 encoded consent string.
    """
    return ConsentV1(base64_decode(consent))



from .bits import Reader
from .iab_tcf import base64_decode, segments, version
from .iab_tcf_v1 import ConsentV1, decode_v1
from .iab_tcf_v2 import ConsentV2, PubRestrictionEntry, decode_v2


def decode(consent: str):
    """Generic implementation of a IAB TCF decoder.

    It detects if the consent received is v1.1 or v2 and returns
    the appropriate ConsentV1 or ConsentV2 instance.
    """

    if consent:
        consent_segments = segments(consent)
        consent_version = version(base64_decode(consent_segments[0]))
        if consent_version == 1:
            return decode_v1(consent_segments[0])
        elif consent_version == 2:
            return decode_v2(consent)
        raise Exception(f"Unable to process a consent with version {consent_version}")
    raise Exception("Unable to process an empty consent")



import json
from typing import List, Tuple


class PubRestrictionEntry:

    """Represents a publisher restriction entry that can be used
    to perform several operation on that information.

    :param purpose_id: The purpose this publisher restriction applies to.
    :param restriction_type: The type this restriction applies to.
    :param restrictions_range: The range of publishers this restriction applies to.
    """

    PURPOSE_FLATLY_NOT_ALLOWED = 0
    REQUIRE_CONSENT = 1
    REQUIRE_LEGITIMATE_INTEREST = 2
    UNDEFINED = 3

    def __init__(
        self,
        purpose_id: int,
        restriction_type: str,
        restrictions_range: List[Tuple[int, int]],
    ):
        self.purpose_id = purpose_id
        self.restriction_type = restriction_type
        self.restrictions_range = restrictions_range

    def __repr__(self):
        return json.dumps(
            {
                "purpose": self.purpose_id,
                "restriction_type": self.restriction_type,
                "restrictions_range": self.restrictions_range,
            }
        )

    def is_not_allowed(self):
        """Checks if the restriction type is PURPOSE_FLATLY_NOT_ALLOWED"""
        return self.restriction_type == self.PURPOSE_FLATLY_NOT_ALLOWED

    def is_consent_required(self):
        """Checks if the restriction type is REQUIRE_CONSENT"""
        return self.restriction_type == self.REQUIRE_CONSENT

    def is_legitimate_interest_required(self):
        """Checks if the restriction type is REQUIRE_LEGITIMATE_INTEREST"""
        return self.restriction_type == self.REQUIRE_LEGITIMATE_INTEREST

    def is_in_range(self, publisher: int) -> bool:
        """Checks if the publisher received is affected by this restriction or not"""
        for (start, end) in self.restrictions_range:
            if publisher >= start and publisher <= end:
                return True
        return False

    def is_publisher_restricted(self, publisher: int) -> bool:
        """If the publisher is in the range specified for this restriction
        and the restriction type is PURPOSE_FLATLY_NOT_ALLOWED then the
        publisher is restricted.
        """
        return self.is_not_allowed() and self.is_in_range(publisher)



from datetime import datetime
from typing import Dict, List, Tuple

from bitarray import bitarray
from bitarray.util import ba2int


class Reader:

    """Represents a bit reader that can extract bits sequentially from
    a bytes consent and return a representation in different formats.

    :param consent: The consent to process in bytes.
    """

    def __init__(self, consent: bytes):
        self._pointer = 0
        self._consent = bitarray(endian="big")
        self._consent.frombytes(consent)

    def read_bits(self, n: int) -> bitarray:
        """Reads the number of bits requested from the pointer position
        and automatically advances the pointer of the reader
        for the subsequent read_bits calls.

        :param n: Number of bits to retrieve.
        """
        self._pointer += n
        return self._consent[self._pointer - n : self._pointer]

    def read_bool(self) -> bool:
        """Reads a bit and returns the value as boolean."""
        return self.read_bits(1).all()

    def read_int(self, n: int) -> int:
        """Reads certain number of bits from the pointer position
        and returns the value as integer.

        :param n: Number of bits to retrieve and transform into int.
        """
        bits = self.read_bits(n)
        return ba2int(bits) if bits else 0

    def read_time(self) -> datetime:
        """Reads 36 bits (the length TCF uses for timestamps) and transforms
        the value into a utc datetime object with seconds granularity.
        """
        seconds = int(self.read_int(36) / 10)
        return datetime.utcfromtimestamp(seconds)

    def read_character(self) -> bytes:
        """Reads 6 bits and returns the value as a character starting with A"""
        return bytes([b"A"[0] + self.read_int(6)])

    def read_string(self, n: int) -> bytes:
        """Reads certain number of bits from the pointer position
        and returns the value as string character by character.

        :param n: Number of bits to retrieve and transform into string.
        """
        return b"".join([self.read_character() for _ in range(n)])

    def read_bitfield(self, n: int) -> Dict[int, bool]:
        """Reads certain number of bits from the pointer position
        and returns the value as a dictionary. The key is the bit
        position (starting by 1) and the value is True if the bit is
        1 and False if the bit is 0.

        :param n: Number of bits to retrieve and transform into the
            dictionary. The dictionary will have as many keys as n.
        """
        return {i + 1: self.read_bool() for i in range(n)}

    def read_range(self, n: int) -> List[Tuple[int, int]]:
        """Reads a complex "ranged" type from the reader given
        the number of ranges we are expecting to find.

        For each range:
        - The first bit indicates if it's a range or just a value.
        - If it's a range, the next 16 bits are the start int and the 
            next 16 bits are the end int.
        - If it's not a range, the next 16 bits are the start and the 
            end is the same as start.

        :param n: Number of ranges we are going to process.
        """
        ranges = []
        for _ in range(n):
            is_range = self.read_bool()
            start = self.read_int(16)
            end = start if not is_range else self.read_int(16)
            ranges.append((start, end))
        return ranges



import base64
from typing import List

from .bits import Reader


def segments(consent: str) -> List[str]:
    """Helper to split the core and non core consents."""
    return consent.split(".")


def base64_decode(segment: str) -> bytes:
    """Helper to decode the IAB TCF segments encoded."""
    padding = "=" * (-len(segment) % 4)
    return base64.urlsafe_b64decode(segment + padding)


def version(consent: bytes) -> int:
    """Helper to extract the version from a consent without having
    to wait for the full decoding.
    """
    reader = Reader(consent)
    return reader.read_int(6)



from typing import Dict

from ..bits import Reader
from .publisher_tc import PubTCEntry


class NonCoreSegment:

    """Represents a non core segment that can contain information
    regarding disclosed vendors, allowed vendors or publiser tc.

    :param segment: The segment to process in bytes.
    """

    CORE = 0
    DISCLOSED_VENDORS = 1
    ALLOWED_VENDORS = 2
    PUBLISHER_TC = 3

    def __init__(self, segment: bytes):
        self._reader = Reader(segment)
        self.type = self._reader.read_int(3)

    def is_disclosed_vendors(self):
        """Checks if the non core segment contains information about Disclosed Vendors."""
        return self.type == self.DISCLOSED_VENDORS

    def is_allowed_vendors(self):
        """Checks if the non core segment contains information about Allowed Vendors."""
        return self.type == self.ALLOWED_VENDORS

    def is_publisher_tc(self):
        """Checks if the non core segment contains information about Publisher TC."""
        return self.type == self.PUBLISHER_TC

    def read_vendors(self) -> Dict[int, bool]:
        """If the non core segment is type Disclosed Vendors or Allowed Vendors, this
        method processes its information extracting a map bitfield with
        the vendors enabled.
        """
        max_vendor_id = self._reader.read_int(16)
        is_range_encoding = self._reader.read_bool()
        if is_range_encoding:
            vendors = {}
            num_entries = self._reader.read_int(12)
            for start, end in self._reader.read_range(num_entries):
                vendors = {**vendors, **{i: True for i in range(start, end + 1)}}
            return vendors
        else:
            return self._reader.read_bitfield(max_vendor_id)

    def read_publisher_tc(self) -> PubTCEntry:
        """If the non core segment is type Publisher TC, this
        method processes and extracts the information inside an entry object for
        easy access.
        """
        purposes_consent = self._reader.read_bitfield(24)
        purposes_lit_transparency = self._reader.read_bitfield(24)
        num_custom_purposes = self._reader.read_int(6)
        custom_purposes_consent = self._reader.read_bitfield(num_custom_purposes)
        custom_purposes_lit_transparency = self._reader.read_bitfield(
            num_custom_purposes
        )
        return PubTCEntry(
            purposes_consent=purposes_consent,
            purposes_lit_transparency=purposes_lit_transparency,
            custom_purposes_consent=custom_purposes_consent,
            custom_purposes_lit_transparency=custom_purposes_lit_transparency,
        )



import os
import subprocess
import logging
import re
from time import time
from typing import List
from setuptools import setup, find_packages
from pathlib import Path


release_file = "RELEASE-VERSION"
requirements_file = "requirements.txt"


logging.basicConfig(format="%(asctime)-15s %(clientip)s %(user)-8s %(message)s")
logger = logging.getLogger(__name__)


def get_version_from_envvar() -> str:
    return os.getenv("RELEASE_VERSION")


def get_version_from_file() -> str:
    try:
        return Path(release_file).read_text().strip()
    except Exception:
        logger.exception(f"Unable to load {release_file} file with static version")


def get_version_from_git() -> str:
    try:
        version = subprocess.check_output(["git", "describe", "--dirty"])
        git_version = version.strip().decode()
        return re.sub(r"-.*$", "-dev" + str(int(time())), git_version)
    except Exception as e:
        logger.exception(f"Unable to load version from git {e}")


def write_version_to_file(version: str):
    Path(release_file).write_text(version)


def get_version() -> str:
    envvar_version = get_version_from_envvar()
    file_version = get_version_from_file()
    git_version = get_version_from_git()
    if not envvar_version and not file_version and not git_version:
        raise Exception("Unable to get the package version.")
    if envvar_version and envvar_version != file_version:
        write_version_to_file(envvar_version)
        return envvar_version
    if git_version and git_version != file_version:
        write_version_to_file(git_version)
        return git_version
    return file_version


def get_requirements() -> List[str]:
    return Path(requirements_file).read_text().strip().splitlines()


setup(
    name="iab-tcf",
    description="A Python implementation of the IAB consent strings (v1.1 and v2)",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Gorka Guridi",
    url="https://github.com/gguridi/iab-tcf",
    author_email="gorka.guridi@gmail.com",
    version=get_version(),
    packages=find_packages(),
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
